from test.testbase import DbIsolatedTest, run_as_sync
from graphql.error import GraphQLError, GraphQLSyntaxError

from db.python.layers import ParticipantLayer
from models.models import (
    SampleUpsertInternal,
    ParticipantUpsertInternal,
    SequencingGroupUpsertInternal,
    AssayUpsertInternal,
)
import api.graphql.schema
from metamist.graphql import gql, validate, configure_sync_client


default_assay_meta = {
    'sequencing_type': 'genome',
    'sequencing_technology': 'short-read',
    'sequencing_platform': 'illumina',
}

SINGLE_PARTICIPANT_UPSERT = ParticipantUpsertInternal(
    external_id='Demeter',
    meta={},
    samples=[
        SampleUpsertInternal(
            external_id='sample_id001',
            meta={},
            type='blood',
            sequencing_groups=[
                SequencingGroupUpsertInternal(
                    type='genome',
                    technology='short-read',
                    platform='illumina',
                    assays=[
                        AssayUpsertInternal(
                            type='sequencing',
                            meta={
                                'reads': [
                                    [
                                        {
                                            'basename': 'sample_id001.filename-R1.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R1.fastq.gz',
                                            'size': 111,
                                        },
                                        {
                                            'basename': 'sample_id001.filename-R2.fastq.gz',
                                            'checksum': None,
                                            'class': 'File',
                                            'location': '/path/to/sample_id001.filename-R2.fastq.gz',
                                            'size': 111,
                                        },
                                    ]
                                ],
                                'reads_type': 'fastq',
                                'batch': 'M001',
                                **default_assay_meta,
                            },
                        ),
                    ],
                )
            ],
        )
    ],
)

TEST_QUERY = gql(
    """
query MyQuery($project: String!) {
  project(name: $project) {
    participants {
      id
      samples {
        id
        sequencingGroups {
          id
          assays {
            id
          }
        }
      }
    }
  }
}"""
)


class TestGraphQL(DbIsolatedTest):
    """Test graphql functionality"""

    @run_as_sync
    async def setUp(self) -> None:
        """Setup the tests"""
        super().setUp()
        self.player = ParticipantLayer(self.connection)

    def test_validate_local_schema(self):
        """
        test using the bundled schema file (from regenerateapi.py)
         to make sure people can validate with authentication
        """
        validate(TEST_QUERY, use_local_schema=True)

    def test_validate_provided_schema(self):
        """
        Validate using schema directly from api.graphql.schema
        (strawberry has an as_str() method)
        """
        client = configure_sync_client(
            schema=api.graphql.schema.schema.as_str(), auth_token='FAKE'
        )
        validate(TEST_QUERY, client=client)

    def test_bad_syntax_query(self):
        """Fail on bad syntax"""
        with self.assertRaises(GraphQLSyntaxError):
            gql(
                """
            query MyQuery(badtoken $project: String!) {
                project(name: $project) {
                    name
                }
            }"""
            )

    def test_bad_field_query(self):
        """Fail because the field doesn't exist"""
        # query syntactically validates
        query = gql(
            """
            query MyQuery($project: String!) {
                project(name: $project) {
                    thisFieldDoesntExist
                }
        }"""
        )
        with self.assertRaises(GraphQLError):
            validate(query, use_local_schema=True)

    @run_as_sync
    async def test_basic_graphql_query(self):
        """Test getting the summary for a project"""
        p = (await self.player.upsert_participants([SINGLE_PARTICIPANT_UPSERT]))[0]

        query = """
query MyQuery($project: String!) {
  project(name: $project) {
    participants {
      id
      samples {
        id
        sequencingGroups {
          id
          assays {
            id
          }
        }
      }
    }
  }
}"""
        data = await self.run_graphql_query_async(
            query, variables={'project': self.project_name}
        )
        participants = data['project']['participants']
        self.assertEqual(1, len(participants))
        self.assertEqual(p.id, participants[0]['id'])
        samples = participants[0]['samples']
        self.assertEqual(1, len(samples))

        self.assertEqual(p.samples[0].to_external().id, samples[0]['id'])
        sequencing_groups = samples[0]['sequencingGroups']
        self.assertEqual(1, len(sequencing_groups))
        self.assertEqual(
            p.samples[0].sequencing_groups[0].to_external().id,
            sequencing_groups[0]['id'],
        )
        assays = sequencing_groups[0]['assays']
        self.assertEqual(
            1, len(participants[0]['samples'][0]['sequencingGroups'][0]['assays'])
        )
        self.assertEqual(
            p.samples[0].sequencing_groups[0].assays[0].id, assays[0]['id']
        )



from test.testbase import DbIsolatedTest, run_as_sync

from db.python.layers import SequencingGroupLayer, SampleLayer
from models.models import (
    SequencingGroupUpsertInternal,
    AssayUpsertInternal,
    SampleUpsertInternal,
)


def get_sample_model():
    """
    Get sample model with sequencing-groups, return in a function
    to protect against any mutation to this model
    """
    return SampleUpsertInternal(
        meta={},
        external_id='EX_ID',
        sequencing_groups=[
            SequencingGroupUpsertInternal(
                type='genome',
                technology='short-read',
                platform='ILLUMINA',
                meta={
                    'meta-key': 'meta-value',
                },
                external_ids={},
                assays=[
                    AssayUpsertInternal(
                        type='sequencing',
                        external_ids={},
                        meta={
                            'sequencing_type': 'genome',
                            'sequencing_platform': 'short-read',
                            'sequencing_technology': 'illumina',
                        },
                    )
                ],
            )
        ],
    )


class TestSequencingGroup(DbIsolatedTest):
    """Test sequencing groups business logic"""

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()
        self.sglayer = SequencingGroupLayer(self.connection)
        self.slayer = SampleLayer(self.connection)

    @run_as_sync
    async def test_insert_sequencing_group(self):
        """Test inserting and fetching a sequencing group"""
        sample_to_insert = get_sample_model()
        sample = await self.slayer.upsert_sample(sample_to_insert)
        sg_id = sample.sequencing_groups[0].id
        sg = await self.sglayer.get_sequencing_group_by_id(sg_id)

        inserted_sg = sample_to_insert.sequencing_groups[0]
        self.assertEqual(inserted_sg.id, sg_id)
        self.assertEqual(inserted_sg.type, sg.type)
        self.assertEqual(inserted_sg.technology, sg.technology)
        self.assertEqual(inserted_sg.platform.lower(), sg.platform.lower())
        self.assertDictEqual(inserted_sg.meta, sg.meta)

    # @run_as_sync
    # async def test_update_sequencing_group(self):
    #     sample = await self.slayer.upsert_sample(get_sample_model())
    #
    #     upsert_sg = SequencingGroupUpsertInternal(
    #         id=sample.sequencing_groups[0].id,
    #         type='transcriptome',
    #     )



# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from from metamist.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from metamist.model.analysis import Analysis
from metamist.model.analysis_query_model import AnalysisQueryModel
from metamist.model.analysis_status import AnalysisStatus
from metamist.model.analysis_update_model import AnalysisUpdateModel
from metamist.model.assay_upsert import AssayUpsert
from metamist.model.body_get_assays_by_criteria import BodyGetAssaysByCriteria
from metamist.model.body_get_latest_complete_analysis_for_type_post import BodyGetLatestCompleteAnalysisForTypePost
from metamist.model.body_get_participants import BodyGetParticipants
from metamist.model.body_get_samples import BodyGetSamples
from metamist.model.error_response import ErrorResponse
from metamist.model.export_type import ExportType
from metamist.model.extra_participant_importer_handler import ExtraParticipantImporterHandler
from metamist.model.family_search_response_data import FamilySearchResponseData
from metamist.model.family_update_model import FamilyUpdateModel
from metamist.model.http_validation_error import HTTPValidationError
from metamist.model.meta_search_entity_prefix import MetaSearchEntityPrefix
from metamist.model.nested_assay import NestedAssay
from metamist.model.nested_family import NestedFamily
from metamist.model.nested_participant import NestedParticipant
from metamist.model.nested_sample import NestedSample
from metamist.model.nested_sequencing_group import NestedSequencingGroup
from metamist.model.paging_links import PagingLinks
from metamist.model.participant_search_response_data import ParticipantSearchResponseData
from metamist.model.participant_upsert import ParticipantUpsert
from metamist.model.project import Project
from metamist.model.project_summary_response import ProjectSummaryResponse
from metamist.model.sample_search_response_data import SampleSearchResponseData
from metamist.model.sample_upsert import SampleUpsert
from metamist.model.search_item import SearchItem
from metamist.model.search_response import SearchResponse
from metamist.model.search_response_model import SearchResponseModel
from metamist.model.search_response_type import SearchResponseType
from metamist.model.sequencing_group_upsert import SequencingGroupUpsert
from metamist.model.validation_error import ValidationError
from metamist.model.web_project import WebProject
